from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from mipac.core.models.drive import RawFile, RawFolder, RawProperties
from mipac.core.models.user import RawUser
from mipac.models.user import User

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions

__all__ = ['Properties', 'File', 'Folder']


class Properties:
    def __init__(self, raw_data: RawProperties) -> None:
        self.__raw_data: RawProperties = raw_data

    @property
    def width(self) -> Optional[int]:
        return self.__raw_data.width

    @property
    def height(self) -> int:
        return self.__raw_data.height

    @property
    def avg_color(self) -> Optional[str]:
        return self.__raw_data.avg_color


class Folder:
    def __init__(self, raw_data: RawFolder):
        self.__raw_data = raw_data

    @property
    def id(self):
        return self.__raw_data.id

    @property
    def created_at(self):
        return self.__raw_data.created_at

    @property
    def name(self):
        return self.__raw_data.name

    @property
    def folders_count(self):
        return self.__raw_data.folders_count

    @property
    def parent_id(self):
        return self.__raw_data.parent_id

    @property
    def parent(self):
        return self.__raw_data.parent

    # @property TODO: 治す
    # def action(self) -> FolderManager:
    #     return mi.framework.manager.ClientActions().drive.get_folder_instance(self.id).action


class File:
    def __init__(self, raw_data: RawFile, *, client: ClientActions):
        self.__raw_data = raw_data
        self.__client: ClientActions = client

    @property
    def id(self):
        return self.__raw_data.id

    @property
    def created_at(self):
        return self.__raw_data.created_at

    @property
    def name(self):
        return self.__raw_data.name

    @property
    def type(self):
        return self.__raw_data.type

    @property
    def md5(self):
        return self.__raw_data.md5

    @property
    def size(self):
        return self.__raw_data.size

    @property
    def is_sensitive(self):
        return self.__raw_data.is_sensitive

    @property
    def blurhash(self):
        return self.__raw_data.blurhash

    @property
    def properties(self):
        return self.__raw_data.properties

    @property
    def url(self):
        return self.__raw_data.url

    @property
    def thumbnail_url(self):
        return self.__raw_data.thumbnail_url

    @property
    def comment(self):
        return self.__raw_data.comment

    @property
    def folder_id(self):
        return self.__raw_data.folder_id

    @property
    def folder(self):
        return self.__raw_data.folder

    @property
    def user_id(self):
        return self.__raw_data.user_id

    @property
    def user(self):
        return User(RawUser(self.__raw_data.user), client=self.__client)



from .chat import Chat
from .drive import File, Folder, Properties
from .emoji import Emoji
from .instance import Instance, InstanceMeta
from .note import Follow, Header, Note, NoteReaction, Poll, Reaction, Renote
from .user import Followee, FollowRequest, User

__all__ = (
    'Chat',
    'Properties',
    'File',
    'Folder',
    'InstanceMeta',
    'Instance',
    'Emoji',
    'User',
    'FollowRequest',
    'Followee',
    'Note',
    'Poll',
    'Reaction',
    'Follow',
    'Header',
    'Renote',
    'NoteReaction',
)



__all__ = (
    'ActiveUsersChartPayload',
    'DriveLocalChartPayload',
    'DriveRemoteChartPayload',
    'DriveChartPayload',
)

from typing import List, TypedDict, Union


class ActiveUsersChartPayload(TypedDict):
    read_write: List[int]
    read: List[int]
    write: List[int]
    registered_within_week: List[int]
    registered_within_month: List[int]
    registered_within_year: List[int]
    registered_outside_week: List[int]
    registered_outside_month: List[int]
    registered_outside_year: List[int]


class DriveLocalChartPayload(TypedDict):
    total_count: List[int]
    total_size: List[int]
    inc_count: List[int]
    inc_size: List[int]
    dec_count: List[int]
    dec_size: List[int]


class DriveRemoteChartPayload(TypedDict):
    total_count: List[int]
    total_size: List[int]
    inc_count: List[int]
    inc_size: List[int]
    dec_count: List[int]
    dec_size: List[int]


class DriveChartPayload(TypedDict):
    local: DriveLocalChartPayload
    remote: DriveRemoteChartPayload



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.actions.admin import AdminActions
from mipac.http import HTTPClient
from mipac.manager.chart import ChartManager
from mipac.manager.chat import ChatManager
from mipac.manager.drive import DriveManager
from mipac.manager.note import NoteManager
from mipac.manager.reaction import ReactionManager
from mipac.manager.user import UserManager

if TYPE_CHECKING:
    from mipac.models.user import User


class ClientActions:
    def __init__(self, session: HTTPClient):
        self.__session: HTTPClient = session
        self.note: NoteManager = NoteManager(session=session, client=self)
        self.chat: ChatManager = ChatManager(session=session, client=self)
        self.user: UserManager = UserManager(session=session, client=self)
        self.admin: AdminActions = AdminActions(
            session=session, client=self
        )  # TODO: 作る
        self.drive: DriveManager = DriveManager(
            session=session, client=self
        )  # TODO: 作る
        self.reaction: ReactionManager = ReactionManager(
            session=session, client=self
        )  # TODO: 作る
        self.chart: ChartManager = ChartManager(
            session=session, client=self
        )  # TODO: 作る

    def _create_user_instance(self, user: User) -> UserManager:
        return UserManager(session=self.__session, client=self, user=user)

    def _create_note_instance(self, note_id: str) -> NoteManager:
        return NoteManager(note_id, session=self.__session, client=self)

    async def get_me(self) -> User:
        return await self.user.action.get_me()



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Optional

from mipac.types import FilePayload, FolderPayload, PropertiesPayload
from mipac.types.user import UserPayload

__all__ = ('RawProperties', 'RawFolder', 'RawFile')


class RawProperties:
    """
    Attributes
    ----------
    width : int
        ファイルの幅
    height : int
        ファイルの高さ
    avg_color : Optional[str]
        ファイルの平均色
    """

    __slots__ = ('width', 'height', 'avg_color')

    def __init__(self, data: PropertiesPayload):
        self.width: Optional[int] = data.get('width')
        self.height: int = data['height']
        self.avg_color: Optional[str] = data.get('avg_color')


class RawFolder:
    """
    Attributes
    ----------
    id : str
        フォルダーのID
    created_at : datetime
        フォルダーの作成された日時
    name : str
        フォルダーの名前
    folders_count : Optional[int]
        # TODO: 調査
    parent_id : str
        親フォルダーのID
    parent : Optional[Dict[str, Any]]
        親フォルダー
    """

    __slots__ = (
        'id',
        'created_at',
        'name',
        'folders_count',
        'parent_id',
        'parent',
    )

    def __init__(self, data: FolderPayload):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.name: str = data['name']
        self.folders_count: Optional[int] = data.get('folders_count', 0)
        self.parent_id: str = data['parent_id']
        self.parent: Optional[Dict[str, Any]] = data.get('parent')


class RawFile:
    """
    Attributes
    ----------
    id : str
        ファイルのID
    created_at : datetime
        ファイルのの作成された日時
    name : str
        ファイルの名前
    type : str
        ファイルの拡張子
    md5 : str
        ファイルのmd5
    size : int
        ファイルのサイズ
    is_sensitive : bool
        このファイルはセンシティブか否か
    blurhash : str
        このファイルのblurhash
    properties : Optional[RawProperties]
        ファイルの情報
    url : str
        ファイルのurl
    thumbnail_url : str
        ファイルのサムネイルurl
    comment : str
        ファイルのコメント
    folder_id : str
        親フォルダのID
    folder : Optional[RawFolder]
        親フォルダの情報？
        # TODO: 調査
    user_id : str
        ファイル作成者のID
    user : Dict[str, Any]
        ファイル作成者の情報
    """

    __slots__ = (
        'id',
        'created_at',
        'name',
        'type',
        'md5',
        'size',
        'is_sensitive',
        'blurhash',
        'properties',
        'url',
        'thumbnail_url',
        'comment',
        'folder_id',
        'folder',
        'user_id',
        'user',
    )

    def __init__(self, data: FilePayload):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.name: str = data['name']
        self.type: str = data['type']
        self.md5: str = data['md5']
        self.size: int = data['size']
        self.is_sensitive: bool = data['is_sensitive']
        self.blurhash: str = data['blurhash']
        self.properties: Optional[RawProperties] = RawProperties(
            data['properties']
        ) if len(data.get('properties')) else None
        self.url: str = data['url']
        self.thumbnail_url: str = data['thumbnail_url']
        self.comment: str = data['comment']
        self.folder_id: str = data['folder_id']
        self.folder: Optional[RawFolder] = RawFolder(
            data['folder']
        ) if data.get('folder') else None
        self.user_id: str = data['user_id']
        self.user: UserPayload = data['user']



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac.core.models.user import RawUser
from mipac.exception import NotSupportedError
from mipac.http import HTTPClient, Route
from mipac.models.user import User

if TYPE_CHECKING:
    from mipac.client import ClientActions


class AdminUserManager:
    def __init__(
        self,
        user_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__user_id = user_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def create_account(self, username: str, password: str) -> User:
        """
        Create a new account.

        Parameters
        ----------
        username : str
            User's name
        password : str
            User's password

        Returns
        -------
        User
            Created user
        """

        if config.is_ayuskey:
            raise NotSupportedError('Ayuskeyではサポートされていません')
        data = {'username': username, 'password': password}
        res = await self.__session.request(
            Route('POST', '/api/admin/accounts/create'),
            json=data,
            auth=True,
            lower=True,
        )
        return User(RawUser(res), client=self.__client)

    async def delete_account(self, user_id: Optional[str] = None) -> bool:
        """
        Deletes the user with the specified user ID.

        Parameters
        ----------
        user_id : Optional[str], default=None
            ID of the user to be deleted
        Returns
        -------
        bool
            Success or failure
        """

        user_id = user_id or self.__user_id

        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/admin/accounts/delete'),
            json=data,
            auth=True,
            lower=True,
        )
        return bool(res)

    async def show_user(self, user_id: Optional[str] = None) -> User:
        """
        Shows the user with the specified user ID.

        Parameters
        ----------
        user_id : Optional[str], default=None
            ID of the user to be shown

        Returns
        -------
        User
        """

        user_id = user_id or self.__user_id
        data = {'userId': user_id}
        res = await self.__session.request(
            Route('GET', '/api/admin/show-user'),
            json=data,
            auth=True,
            lower=True,
        )
        return User(RawUser(res), client=self.__client)

    async def suspend(self, user_id: Optional[str] = None) -> bool:
        """
        Suspends the user with the specified user ID.

        Parameters
        ----------
        user_id : Optional[str], default=None
            ID of the user to be suspended

        Returns
        -------
        bool
            Success or failure
        """

        user_id = user_id or self.__user_id
        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/admin/suspend-user'),
            json=data,
            auth=True,
            lower=True,
        )
        return bool(res)

    async def unsuspend(self, user_id: Optional[str] = None) -> bool:
        """
        Unsuspends the user with the specified user ID.

        Parameters
        ----------
        user_id : Optional[str], default=None
            ID of the user to be unsuspended

        Returns
        -------
        bool
            Success or failure
        """

        user_id = user_id or self.__user_id
        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/admin/unsuspend-user'),
            json=data,
            auth=True,
            lower=True,
        )
        return bool(res)



from typing import List, Optional, TypedDict

__all__ = ('EmojiPayload',)


class EmojiPayload(TypedDict):
    id: Optional[str]
    aliases: Optional[List[str]]
    name: Optional[str]
    category: Optional[str]
    host: Optional[str]
    url: Optional[str]



from __future__ import annotations

from datetime import datetime

from mipac.types import NoteReactionPayload, UserPayload

__all__ = ['RawNoteReaction']


class RawNoteReaction:
    """
    Attributes
    ----------
    id : str
        TODO: 調査
    created_at : datetime
        リアクションが付けられた時間
    user : UserPayload
        リアクションを付けたユーザー
    reaction : str
    """

    __slots__ = ('id', 'created_at', 'user', 'reaction')

    def __init__(self, data: NoteReactionPayload):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.user: UserPayload = data['user']
        self.reaction: str = data['type']



from __future__ import annotations

from typing import Any, Dict, List, Optional, TypedDict

from .drive import FilePayload
from .emoji import EmojiPayload
from .instance import InstancePayload

__all__ = (
    'ChannelPayload',
    'FieldContentPayload',
    'UserPayload',
    'PinnedPagePayload',
    'PinnedNotePayload',
    'OptionalUser',
)


class ChannelPayload(TypedDict):
    id: Optional[str]
    created_at: Optional[str]
    last_noted_at: Optional[str]
    name: Optional[str]
    description: Optional[str]
    banner_url: Optional[str]
    notes_count: Optional[int]
    users_count: Optional[int]
    is_following: Optional[bool]
    user_id: Optional[str]


class PinnedNotePayload(TypedDict):
    id: Optional[str]
    created_at: Optional[str]
    text: Optional[str]
    cw: Optional[str]
    user_id: Optional[str]
    user: Optional['UserPayload']
    reply_id: Optional[str]
    renote_id: Optional[str]
    reply: Optional[Dict[str, Any]]
    renote: Optional[Dict[str, Any]]
    via_mobile: Optional[bool]
    is_hidden: Optional[bool]
    visibility: Optional[str]
    mentions: Optional[List[str]]
    visible_user_ids: Optional[List[str]]
    file_ids: Optional[List[str]]
    files: Optional[List[FilePayload]]
    tags: Optional[List[str]]
    poll: Optional[Dict[str, Any]]
    channel_id: Optional[str]
    channel: Optional[ChannelPayload]
    local_only: Optional[bool]
    emojis: Optional[List[EmojiPayload]]
    reactions: Optional[Dict[str, Any]]
    renote_count: Optional[int]
    replies_count: Optional[int]
    uri: Optional[str]
    url: Optional[str]
    my_reaction: Optional[Dict[str, Any]]


class PinnedPagePayload(TypedDict):
    id: Optional[str]
    created_at: Optional[str]
    updated_at: Optional[str]
    title: Optional[str]
    name: Optional[str]
    summary: Optional[str]
    content: Optional[List]
    variables: Optional[List]
    user_id: Optional[str]
    author: Optional[Dict[str, Any]]


class FieldContentPayload(TypedDict):
    name: str
    value: str


class OptionalUser(TypedDict, total=False):
    user_id: str
    name: str
    host: str
    is_admin: bool
    is_moderator: bool
    is_bot: bool
    is_cat: bool
    is_lady: bool
    online_status: str


class UserPayload(OptionalUser):
    id: str
    username: str
    avatar_url: Optional[str]
    avatar_blurhash: Optional[str]
    avatar_color: Optional[str]
    emojis: Optional[List[str]]
    url: str
    uri: str
    created_at: str
    updated_at: str
    is_locked: bool
    is_silenced: bool
    is_suspended: bool
    description: str
    location: str
    birthday: str
    fields: Any
    followers_count: int
    following_count: int
    notes_count: int
    pinned_note_ids: List[str]
    pinned_notes: List[str]
    pinned_page_id: str
    pinned_page: str
    ff_visibility: str
    is_following: bool
    is_follow: bool
    is_blocking: bool
    is_blocked: bool
    is_muted: bool
    instance: Optional[InstancePayload]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from mipac import AbstractManager
from mipac.core.models.drive import RawFile, RawFolder
from mipac.exception import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.drive import File, Folder
from mipac.util import remove_dict_empty

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ['FolderManager', 'FileManager', 'DriveManager']


class FileManager:
    def __init__(
        self,
        file_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.__file_id = file_id

    async def show_file(
        self, file_id: Optional[str], url: Optional[str]
    ) -> File:
        """
        ファイルの情報を取得します。

        Parameters
        ----------
        file_id : Optional[str], default=None
            ファイルのID
        url : Optional[str], default=None
            ファイルのURL

        Returns
        -------
        File
            ファイルの情報
        """

        data = remove_dict_empty({'fileId': file_id, 'url': url})
        res = await self.__session.request(
            Route('POST', '/api/admin/drive/show-file'),
            json=data,
            auth=True,
            lower=True,
        )
        return File(RawFile(res), client=self.__client)

    async def remove_file(self, file_id: Optional[str] = None) -> bool:
        """
        指定したIDのファイルを削除します

        Parameters
        ----------
        file_id : Optional[str], default=None
            削除するファイルのID

        Returns
        -------
        bool
            削除に成功したかどうか
        """

        file_id = file_id or self.__file_id
        return bool(
            await self.__session.request(
                Route('POST', '/api/drive/files/delete'),
                json={'fileId': file_id},
                auth=True,
            )
        )

    async def get_files(
        self,
        limit: int = 10,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        folder_id: Optional[str] = None,
        file_type: Optional[str] = None,
    ) -> List[File]:
        """
        ファイルを取得します

        Parameters
        ----------
        limit : int, default=10
            取得する上限
        since_id : Optional[str], default=None
            指定すると、そのIDを起点としてより新しいファイルを取得します
        until_id : Optional[str], default=None
            指定すると、そのIDを起点としてより古いファイルを取得します
        folder_id : Optional[str], default=None
            指定すると、そのフォルダーを起点としてファイルを取得します
        file_type : Optional[str], default=None
            取得したいファイルの拡張子
        """
        if limit > 100:
            raise ParameterError('limit must be less than 100')

        data = {
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'folderId': folder_id,
            'Type': file_type,
        }
        res = await self.__session.request(
            Route('POST', '/api/drive/files'), json=data, auth=True, lower=True
        )
        return [File(RawFile(i), client=self.__client) for i in res]


class FolderManager(AbstractManager):
    def __init__(
        self,
        folder_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__folder_id = folder_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def create(self, name: str, parent_id: Optional[str] = None) -> bool:
        """
        フォルダーを作成します

        Parameters
        ----------
        name : str, default=None
            フォルダーの名前
        parent_id : Optional[str], default=None
            親フォルダーのID

        Returns
        -------
        bool
            作成に成功したか否か
        """

        parent_id = parent_id or self.__folder_id

        data = {'name': name, 'parent_id': parent_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/drive/folders/create'),
                json=data,
                lower=True,
                auth=True,
            )
        )

    async def delete(self, folder_id: Optional[str] = None) -> bool:
        """
        Parameters
        ----------
        folder_id : Optional[str] = None
            削除するノートのID

        Returns
        -------
        bool
            削除に成功したか否か
        """

        folder_id = folder_id or self.__folder_id
        data = {'folderId': folder_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/drive/folders/delete'),
                json=data,
                lower=True,
                auth=True,
            )
        )

    async def get_files(
        self,
        limit: int = 10,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        folder_id: Optional[str] = None,
        file_type: Optional[str] = None,
    ) -> List[File]:
        """
        ファイルを取得します

        Parameters
        ----------
        limit : int, default=10
            取得する上限
        since_id : Optional[str], default=None
            指定すると、そのIDを起点としてより新しいファイルを取得します
        until_id : Optional[str], default=None
            指定すると、そのIDを起点としてより古いファイルを取得します
        folder_id : Optional[str], default=None
            指定すると、そのフォルダーを起点としてファイルを取得します
        file_type : Optional[str], default=None
            取得したいファイルの拡張子
        """
        if limit > 100:
            raise ParameterError('limit must be less than 100')

        folder_id = folder_id or self.__folder_id
        data = {
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'folderId': folder_id,
            'Type': file_type,
        }
        res = await self.__session.request(
            Route('POST', '/api/drive/files'), json=data, auth=True, lower=True
        )
        return [File(RawFile(i), client=self.__client) for i in res]


class DriveManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def get_folders(
        self,
        limit: int = 100,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        folder_id: Optional[str] = None,
    ) -> List[Folder]:
        """
        フォルダーの一覧を取得します

        Parameters
        ----------
        limit : int, default=10
            取得する上限
        since_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより新しい投稿を取得します
        until_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより古い投稿を取得します
        folder_id : Optional[str], default=None
            指定すると、そのフォルダーを起点としてフォルダーを取得します
        """

        data = {
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'folderId': folder_id,
        }
        data = await self.__session.request(
            Route('POST', '/api/drive/folders'),
            json=data,
            lower=True,
            auth=True,
        )
        return [Folder(RawFolder(i)) for i in data]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, TypedDict

if TYPE_CHECKING:
    from mipac.types import PropertiesPayload, UserPayload


class PageContentPayload(TypedDict):
    id: str
    type: str
    text: Optional[str]
    file_id: Optional[str]
    width: Optional[int]
    height: Optional[int]
    note: Optional[str]
    detailed: Optional[bool]
    fn: Optional[Any]
    var: Optional[Any]
    event: Optional[Any]
    action: Optional[str]
    content: Optional[str]
    message: Optional[Any]
    primary: Optional[bool]
    inc: Optional[int]
    canvas_id: Optional[str]
    attach_canvas_image: Optional[bool]
    default: Optional[str]
    value: Optional[List[Any]]

    children: Optional['PageContentPayload']


class VariablePayload(TypedDict):
    id: str
    name: str
    type: str
    value: Optional[str]


class PageFilePayload(TypedDict):
    id: str
    created_at: str
    name: str
    type: str
    md5: str
    size: int
    is_sensitive: bool
    blurhash: str
    properties: PropertiesPayload
    url: str
    thumbnail_url: str
    comment: Optional[str]
    folder_id: Optional[str]
    folder: Any
    user_id: str
    user: Any


class EyeCatchingImagePayload(PageFilePayload):
    pass


class AttachedFilePayload(PageFilePayload):
    pass


class PagePayload(TypedDict):
    id: str
    created_at: str
    updated_at: str
    user_id: str
    user: UserPayload
    content: List[PageContentPayload]
    variable: List[VariablePayload]
    title: str
    name: str
    summary: Optional[str]
    hide_title_when_pinned: bool
    align_center: bool
    font: str
    script: str
    eye_catching_image_id: Optional[str]
    eye_catching_image: EyeCatchingImagePayload
    attached_files: List[AttachedFilePayload]
    liked_count: int



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from mipac.core.models.note import RawNote
from mipac.exception import ParameterError
from mipac.http import HTTPClient, Route
from mipac.manager.favorite import FavoriteManager
from mipac.manager.file import MiFile
from mipac.manager.reaction import ReactionManager

__all__ = ['NoteActions']

from mipac.models.note import Note, NoteReaction, Poll
from mipac.util import check_multi_arg, remove_dict_empty

if TYPE_CHECKING:
    from mipac.client import ClientActions


class NoteActions:
    def __init__(
        self,
        note_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__note_id: Optional[str] = note_id
        self.favorite = FavoriteManager(
            note_id=note_id, session=session, client=client
        )
        self.reaction = ReactionManager(
            note_id=note_id, session=session, client=client
        )
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add_clips(
        self, clip_id: str, note_id: Optional[str] = None
    ) -> bool:
        """
        クリップに追加します

        Parameters
        ----------
        note_id : Optional[str], default=None
                追加するノートのID
        clip_id : str
            クリップのID
        note_id : Optional[str], default=None
            追加したいノートのID

        Returns
        -------
        bool
            成功したか否か
        """

        note_id = note_id or self.__note_id

        data = {'noteId': note_id, 'clipId': clip_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/clips/add-note'), json=data, auth=True
            )
        )

    async def send(
        self,
        content: Optional[str] = None,
        visibility: str = 'public',
        visible_user_ids: Optional[List[str]] = None,
        cw: Optional[str] = None,
        local_only: bool = False,
        extract_mentions: bool = True,
        extract_hashtags: bool = True,
        extract_emojis: bool = True,
        reply_id: Optional[str] = None,
        renote_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        files: Optional[List[MiFile]] = None,
        poll: Optional[Poll] = None,
    ) -> Note:
        """
        ノートを投稿します。

        Parameters
        ----------
        content : Optional[str], default=None
            投稿する内容
        visibility : str, optional
            公開範囲, by default "public"
            Enum: "public" "home" "followers" "specified"
        visible_user_ids : Optional[List[str]], optional
            公開するユーザー, by default None
        cw : Optional[str], optional
            閲覧注意の文字, by default None
        local_only : bool, optional
            ローカルにのみ表示するか, by default False
        extract_mentions : bool, optional
            メンションを展開するか, by default False
        extract_hashtags : bool, optional
            ハッシュタグを展開するか, by default False
        extract_emojis : bool, optional
            絵文字を展開するか, by default False
        reply_id : Optional[str], optional
            リプライ先のid, by default None
        renote_id : Optional[str], optional
            リノート先のid, by default None
        channel_id : Optional[str], optional
            チャンネルid, by default None
        files : List[MiFile], optional
            添付するファイルのリスト, by default None
        poll : Optional[Poll], optional
            アンケート, by default None

        Returns
        -------
        Note
            投稿したノート

        Raises
        ------
        ContentRequired
            [description]
        """

        field = {
            'visibility': visibility,
            'visibleUserIds': visible_user_ids,
            'text': content,
            'cw': cw,
            'localOnly': local_only,
            'noExtractMentions': not extract_mentions,
            'noExtractHashtags': not extract_hashtags,
            'noExtractEmojis': not extract_emojis,
            'replyId': reply_id,
            'renoteId': renote_id,
            'channelId': channel_id,
        }
        if not check_multi_arg(content, files, renote_id, poll):
            raise ParameterError(
                'ノートの送信にはcontent, file_ids, renote_id またはpollのいずれか1つが無くてはいけません'
            )

        if poll and type(Poll):
            poll_data = remove_dict_empty(
                {
                    'choices': poll.choices,
                    'multiple': poll.multiple,
                    'expiresAt': poll.expires_at,
                    'expiredAfter': poll.expired_after,
                }
            )
            field['poll'] = poll_data

        # if files:  # TODO: get_file_idsを直さないと使えない
        # field['fileIds'] = await get_file_ids(files=files)
        field = remove_dict_empty(field)
        res = await self.__session.request(
            Route('POST', '/api/notes/create'),
            json=field,
            auth=True,
            lower=True,
        )
        return Note(RawNote(res['created_note']), client=self.__client)

    async def delete(self, note_id: Optional[str] = None) -> bool:
        """
        ノートを削除します

        Parameters
        ----------
        note_id : Optional[str], default=None
            削除したいノートのID

        Returns
        -------
        bool
            削除に成功したか否か
        """

        note_id = note_id or self.__note_id

        data = {'noteId': note_id}
        res = await self.__session.request(
            Route('POST', '/api/notes/delete'), json=data, auth=True
        )
        return bool(res)

    async def create_renote(self, note_id: Optional[str] = None) -> Note:
        """
        リノートを作成します

        Parameters
        ----------
        note_id : Optional[str], default=None
            ノートのID

        Returns
        -------
        Note
            作成したリノート
        """
        note_id = note_id or self.__note_id
        return await self.send(renote_id=note_id)

    async def create_quote(
        self,
        content: Optional[str] = None,
        visibility: str = 'public',
        visible_user_ids: Optional[List[str]] = None,
        cw: Optional[str] = None,
        local_only: bool = False,
        extract_mentions: bool = True,
        extract_hashtags: bool = True,
        extract_emojis: bool = True,
        file_ids: Optional[List[str]] = None,
        poll: Optional[Poll] = None,
        note_id: Optional[str] = None,
    ) -> Note:
        """
        Create a note quote.

        Parameters
        ----------
        content: Optional[str], default=None
            text
        visibility: str, default='public'
            Disclosure range
        visible_user_ids: Optional[List[str]], default=None
            List of users to be published
        cw: Optional[str], default=None
            Text to be displayed when warning is given
        local_only: bool, default=False
            Whether to show only locally or not
        extract_mentions: bool, default=True
            Whether to expand the mention
        extract_hashtags: bool, default=True
            Whether to expand the hashtag
        extract_emojis: bool, default=True
            Whether to expand the emojis
        file_ids: Optional[List[str]], default=None
            The ID list of files to be attached
        poll: Optional[Poll], default=None
            Questionnaire to be created
        note_id: Optional[str], default=None
            Note IDs to target for renote and citations
        """

        note_id = note_id or self.__note_id

        return await self.send(
            content=content,
            visibility=visibility,
            visible_user_ids=visible_user_ids,
            cw=cw,
            local_only=local_only,
            extract_mentions=extract_mentions,
            extract_hashtags=extract_hashtags,
            extract_emojis=extract_emojis,
            renote_id=note_id,
            poll=poll,
        )  # TODO: filesを受け取るように

    async def get_note(self, note_id: Optional[str] = None) -> Note:
        """
        ノートを取得します

        Parameters
        ----------
        note_id : Optional[str], default=None
            ノートのID

        Returns
        -------
        Note
            取得したノートID
        """
        note_id = note_id or self.__note_id
        res = await self.__session.request(
            Route('POST', '/api/notes/show'),
            json={'noteId': note_id},
            auth=True,
            lower=True,
        )
        return Note(RawNote(res), client=self.__client)

    async def get_replies(
        self,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        limit: int = 10,
        note_id: Optional[str] = None,
    ) -> List[Note]:
        """
        ノートに対する返信を取得します

        Parameters
        ---------
        since_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより新しい投稿を取得します
        until_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより古い投稿を取得します
        limit : int, default=10
            取得する上限
        note_id: Optional[str], default=None
            返信を取得したいノートのID

        Returns
        -------
        List[Note]
            返信のリスト
        """
        note_id = note_id or self.__note_id
        res = await self.__session.request(
            Route('POST', '/api/notes/replies'),
            json={
                'noteId': note_id,
                'sinceId': since_id,
                'untilId': until_id,
                'limit': limit,
            },
            auth=True,
            lower=True,
        )
        return [Note(RawNote(i), client=self.__client) for i in res]

    async def get_reaction(
        self, reaction: str, note_id: Optional[str] = None
    ) -> List[NoteReaction]:
        note_id = note_id or self.__note_id
        # return await ReactionManager(note_id=note_id, client=self.__client).get_reaction(reaction)
        # TODO:  self.__client形式に置き換え



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.core.models.chat import RawChat

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions

__all__ = ['Chat']


class Chat:
    """
    チャットオブジェクト
    """

    def __init__(self, raw_data: RawChat, *, client: ClientActions):
        self.__raw_data: RawChat = raw_data
        self.__client: ClientActions = client

    @property
    def id(self):
        return self.__raw_data.id

    @property
    def created_at(self):
        return self.__raw_data.created_at

    @property
    def content(self):
        return self.__raw_data.content

    @property
    def user_id(self):
        return self.__raw_data.user_id

    @property
    def author(self):
        return self.__raw_data.author

    @property
    def recipient_id(self):
        return self.__raw_data.recipient_id

    @property
    def recipient(self):
        return self.__raw_data.recipient

    @property
    def group_id(self):
        return self.__raw_data.group_id

    @property
    def file_id(self):
        return self.__raw_data.file_id

    @property
    def is_read(self):
        return self.__raw_data.is_read

    @property
    def reads(self):
        return self.__raw_data.reads

    async def delete(self) -> bool:
        """
        チャットを削除します（チャットの作者である必要があります）

        Returns
        -------
        bool:
            成功したか否か
        """
        res = await self.__client.chat.delete(message_id=self.id)
        return bool(res)



"""
MiPAを使用する上でちょっとした際に便利なツール一覧
"""
from __future__ import annotations

import re
from datetime import datetime, timedelta
from inspect import isawaitable
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar

__all__ = (
    'deprecated_func',
    'MiTime',
    'get_cache_key',
    'key_builder',
    'check_multi_arg',
    'async_all',
    'find',
    'remove_list_empty',
    'remove_dict_empty',
    'upper_to_lower',
    'str_lower',
    'bool_to_string',
)

T = TypeVar('T')


def deprecated_func(func):
    print(f'deprecated function:{func.__name__}')


class MiTime:
    def __init__(self, start: timedelta, end: datetime):
        self.start = start
        self.end = end


def get_cache_key(func):
    async def decorator(self, *args, **kwargs):
        ordered_kwargs = sorted(kwargs.items())
        key = (
            (func.__module__ or '')
            + '.{0}'
            + f'{self}'
            + str(args)
            + str(ordered_kwargs)
        )
        return await func(self, *args, **kwargs, cache_key=key)

    return decorator


def key_builder(func, cls, *args, **kwargs):
    ordered_kwargs = sorted(kwargs.items())
    key = (
        (func.__module__ or '')
        + f'.{func.__name__}'
        + f'{cls}'
        + str(args)
        + str(ordered_kwargs)
    )
    return key


def check_multi_arg(*args: Any) -> bool:
    """
    複数の値を受け取り値が存在するかをboolで返します

    Parameters
    ----------
    args : list
        確認したい変数のリスト

    Returns
    -------
    bool
        存在する場合はTrue, 存在しない場合はFalse
    """
    return bool([i for i in args if i])


async def async_all(gen, *, check=isawaitable):
    for elem in gen:
        if check(elem):
            elem = await elem
        if not elem:
            return False
    return True


def find(predicate: Callable[[T], Any], seq: Iterable[T]) -> Optional[T]:
    """A helper to return the first element found in the sequence
    that meets the predicate. For example: ::

        member = discord.utils.find(lambda m: m.name == 'Mighty', channel.guild.members)

    would find the first :class:`~discord.Member` whose name is 'Mighty' and return it.
    If an entry is not found, then ``None`` is returned.

    This is different from :func:`py:filter` due to the fact it stops the moment it finds
    a valid entry.

    Parameters
    -----------
    predicate
        A function that returns a boolean-like result.
    seq: :class:`collections.abc.Iterable`
        The iterable to search through.
    """
    for element in seq:
        if predicate(element):
            return element
    return None


def remove_list_empty(data: List[Any]) -> List[Any]:
    """
    Parameters
    ----------
    data: dict
        空のkeyを削除したいdict

    Returns
    -------
    Dict[str, Any]
        空のkeyがなくなったdict
    """
    return [k for k in data if k]


def remove_dict_empty(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Parameters
    ----------
    data: dict
        空のkeyを削除したいdict

    Returns
    -------
    _data: dict
        空のkeyがなくなったdict
    """

    _data = {}
    _data = {k: v for k, v in data.items() if v is not None}
    _data = {k: v for k, v in data.items() if v}
    return _data


def upper_to_lower(
    data: Dict[str, Any],
    field: Optional[Dict[str, Any]] = None,
    nest: bool = True,
    replace_list: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """
    Parameters
    ----------
    data: dict
        小文字にしたいkeyがあるdict
    field: dict, default=None
        謎
    nest: bool, default=True
        ネストされたdictのkeyも小文字にするか否か
    replace_list: dict, default=None
        dictのkey名を特定の物に置き換える

    Returns
    -------
    field : dict
        小文字になった, key名が変更されたdict
    """

    if data is None:
        return {}
    if replace_list is None:
        replace_list = {}

    if field is None:
        field = {}
    for attr in data:
        pattern = re.compile('[A-Z]')
        large = [i.group().lower() for i in pattern.finditer(attr)]
        result = [None] * (len(large + pattern.split(attr)))
        result[::2] = pattern.split(attr)
        result[1::2] = ['_' + i.lower() for i in large]
        default_key = ''.join(result)
        if replace_list.get(attr):
            default_key = default_key.replace(attr, replace_list.get(attr))
        field[default_key] = data[attr]
        if isinstance(field[default_key], dict) and nest:
            field[default_key] = upper_to_lower(field[default_key])
    return field


def str_lower(text: str):
    pattern = re.compile('[A-Z]')
    large = [i.group().lower() for i in pattern.finditer(text)]
    result = [None] * (len(large + pattern.split(text)))
    result[::2] = pattern.split(text)
    result[1::2] = ['_' + i.lower() for i in large]
    return ''.join(result)


def bool_to_string(boolean: bool) -> str:
    """
    boolを小文字にして文字列として返します

    Parameters
    ----------
    boolean : bool
        変更したいbool値
    Returns
    -------
    true or false: str
        小文字になったbool文字列
    """
    return 'true' if boolean else 'false'



from __future__ import annotations

from typing import List, Optional

from mipac.types import PollPayload

__all__ = ['RawPollChoices', 'RawPoll']


class RawPollChoices:
    """
    Attributes
    ----------
    text : str
        アンケートの項目名
    votes : int
        投票数
    is_voted : bool
        投票済みか否か
    """

    __slots__ = ('text', 'votes', 'is_voted')

    def __init__(self, data):
        self.text: str = data['text']
        self.votes: int = data['votes']
        self.is_voted: bool = data['isVoted']


class RawPoll:
    """
    Attributes
    ----------
    multiple : Optional[bool]
        複数回投票可能か否か
    expires_at : Optional[int]
        投票期限
    choices : Optional[List[RawPollChoices]]
        項目
    expired_after : Optional[int]
        残り期限
    """

    __slots__ = ('multiple', 'expires_at', 'choices', 'expired_after')

    def __init__(self, data: PollPayload):
        self.multiple: Optional[bool] = data.get('multiple')
        self.expires_at: Optional[int] = data.get('expires_at')
        self.choices: Optional[List[RawPollChoices]] = [
            RawPollChoices(i) for i in data['choices']
        ] if data.get('choices') else None
        self.expired_after: Optional[int] = data.get('expired_after')



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, TypedDict

if TYPE_CHECKING:
    from mipac.types.user import UserPayload


__all__ = ('PropertiesPayload', 'FolderPayload', 'FilePayload')


class PropertiesPayload(TypedDict):
    """
    プロパティー情報
    """

    width: int
    height: int
    avg_color: Optional[str]


class FolderPayload(TypedDict):
    """
    フォルダーの情報
    """

    id: str
    created_at: str
    name: str
    folders_count: int
    files_count: int
    parent_id: str
    parent: Dict[str, Any]


class FilePayload(TypedDict):
    """
    ファイル情報
    """

    id: str
    created_at: str
    name: str
    type: str
    md5: str
    size: int
    is_sensitive: bool
    blurhash: str
    properties: PropertiesPayload
    url: str
    thumbnail_url: str
    comment: str
    folder_id: str
    folder: FolderPayload
    user_id: str
    user: UserPayload  # TODO: ここ確認



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac import AbstractManager
from mipac.core.models.chart import RawActiveUsersChart, RawDriveChart
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ('ChartManager',)


class ChartManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def get_active_user(
        self, span: str = 'day', limit: int = 30, offset: int = 0
    ) -> RawActiveUsersChart:
        data = {'span': span, 'limit': limit, 'offset': offset}
        data = await self.__session.request(
            Route('POST', '/api/charts/active-users'),
            json=data,
            auth=True,
            lower=True,
        )
        return RawActiveUsersChart(data)

    async def get_drive(
        self, span: str = 'day', limit: int = 30, offset: int = 0
    ) -> RawDriveChart:
        data = {'span': span, 'limit': limit, 'offset': offset}
        data = await self.__session.request(
            Route('POST', '/api/charts/drive'),
            json=data,
            auth=True,
            lower=True,
        )
        return RawDriveChart(data)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from mipac import AbstractManager, HTTPClient
from mipac.core.models.emoji import RawEmoji
from mipac.core.models.reaction import RawNoteReaction
from mipac.http import Route
from mipac.models.emoji import Emoji
from mipac.models.note import NoteReaction
from mipac.util import remove_dict_empty

if TYPE_CHECKING:
    from mipac.client import ClientActions


class ReactionManager(AbstractManager):
    def __init__(
        self,
        note_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__note_id: Optional[str] = note_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add(self, reaction: str, note_id: Optional[str] = None) -> bool:
        """
        指定したnoteに指定したリアクションを付与します（内部用

        Parameters
        ----------
        reaction : Optional[str]
            付与するリアクション名
        note_id : Optional[str]
            付与対象のノートID

        Returns
        -------
        bool
            成功したならTrue,失敗ならFalse
        """
        note_id = note_id or self.__note_id

        data = remove_dict_empty({'noteId': note_id, 'reaction': reaction})
        route = Route('POST', '/api/notes/reactions/create')
        return await self.__session.request(
            route, json=data, auth=True, lower=True
        )

    async def remove(self, note_id: Optional[str] = None) -> bool:
        note_id = note_id or self.__note_id

        data = remove_dict_empty({'noteId': note_id})
        route = Route('POST', '/api/notes/reactions/delete')
        return bool(
            await self.__session.request(
                route, json=data, auth=True, lower=True
            )
        )

    async def get_reaction(
        self, reaction: str, note_id: Optional[str] = None, *, limit: int = 11
    ) -> List[NoteReaction]:
        note_id = note_id or self.__note_id
        data = remove_dict_empty(
            {'noteId': note_id, 'limit': limit, 'type': reaction}
        )
        res = await self.__session.request(
            Route('POST', '/api/notes/reactions'),
            json=data,
            auth=True,
            lower=True,
        )
        return [NoteReaction(RawNoteReaction(i)) for i in res]

    async def get_emoji_list(self) -> List[Emoji]:
        data = await self.__session.request(
            Route('GET', '/api/meta'), json={'detail': False}, auth=True
        )
        return [Emoji(RawEmoji(i)) for i in data['emojis']]



from __future__ import annotations

from mipac.core.models.emoji import RawEmoji

__all__ = ('Emoji',)


class Emoji:
    def __init__(self, raw_data: RawEmoji):
        self.__raw_data = raw_data

    @property
    def id(self):
        return self.__raw_data.id

    @property
    def aliases(self):
        return self.__raw_data.aliases

    @property
    def name(self):
        return self.__raw_data.name

    @property
    def category(self):
        return self.__raw_data.category

    @property
    def host(self):
        return self.__raw_data.host

    @property
    def url(self):
        return self.__raw_data.url

    # @property
    # def action(self):
    #     return mi.framework.manager.ClientActions().emoji



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from mipac.types import UserPayload

__all__ = ('NoteReactionPayload',)


class NoteReactionPayload(TypedDict):
    id: str
    created_at: str
    user: UserPayload
    type: str



from mipac.core.models.chart import (RawActiveUsersChart, RawDriveChart,
                                     RawDriveLocalChart, RawDriveRemoteChart)
from mipac.core.models.chat import RawChat
from mipac.core.models.drive import RawFile, RawFolder, RawProperties
from mipac.core.models.emoji import RawEmoji
from mipac.core.models.instance import RawInstance
from mipac.core.models.note import RawNote, RawReaction, RawRenote
from mipac.core.models.poll import RawPoll, RawPollChoices
from mipac.core.models.reaction import RawNoteReaction
from mipac.core.models.user import RawUser, RawUserDetails

__all__ = (
    'RawUserDetails',
    'RawUser',
    'RawActiveUsersChart',
    'RawDriveRemoteChart',
    'RawDriveLocalChart',
    'RawDriveChart',
    'RawNoteReaction',
    'RawPoll',
    'RawPollChoices',
    'RawRenote',
    'RawReaction',
    'RawNote',
    'RawInstance',
    'RawEmoji',
    'RawProperties',
    'RawFolder',
    'RawFile',
    'RawChat',
)



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional

from mipac.core.models.drive import RawFile
from mipac.core.models.emoji import RawEmoji
from mipac.core.models.poll import RawPoll
from mipac.core.models.user import RawUser
from mipac.types import NotePayload, ReactionPayload, RenotePayload
from mipac.util import upper_to_lower

__all__ = ('RawRenote', 'RawReaction', 'RawNote')


class RawRenote:
    """
    Attributes
    ----------
    id : str
    created_at : datetime
    user_id :str
    user : RawUser
    content: Optional[str], default=None
    cw : Optional[str], default=None
    visibility : str
    renote_count : int
    replies_count : int
    reactions
    emojis
    file_ids : List[str]
    files
    reply_id
    renote_id
    uri
    poll Optional[RawPoll]
    """

    __slots__ = (
        'id',
        'created_at',
        'user_id',
        'user',
        'content',
        'cw',
        'visibility',
        'renote_count',
        'replies_count',
        'replies_count',
        'reactions',
        'emojis',
        'file_ids',
        'files',
        'reply_id',
        'renote_id',
        'uri',
        'poll',
    )

    def __init__(self, data: RenotePayload):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.user_id: str = data['user_id']
        self.user: RawUser = RawUser(data['user'])
        self.content: Optional[str] = data.get('text', None)
        self.cw: Optional[str] = data['cw']
        self.visibility: str = data['visibility']
        self.renote_count: int = data['renote_count']
        self.replies_count: int = data['replies_count']
        self.reactions = data['reactions']  # TODO:型探す
        self.emojis = data['emojis']  # TODO:型探す
        self.file_ids: List[str] = data['file_ids']
        self.files = data['files']
        self.reply_id = data['reply_id']
        self.files = data['files']
        self.reply_id = data['reply_id']
        self.renote_id = data['renote_id']
        self.uri = data.get('uri')
        self.poll: Optional[RawPoll] = RawPoll(data['poll']) if data.get(
            'poll'
        ) else None


class RawReaction:
    """
    Attributes
    ----------
    id : Optional[str], default=None
    created_at : Optional[datetime], default=None
    type : Optional[str], default=None
    is_read : bool
    user : Optional[RawUser], default=None
    note : Optional[RawNote], default=None
    reaction : str
    """

    __slots__ = (
        'id',
        'created_at',
        'type',
        'is_read',
        'user',
        'note',
        'reaction',
    )

    def __init__(self, data: ReactionPayload):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[str] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.type: Optional[str] = data.get('type')
        self.is_read: bool = bool(data.get('is_read'))
        self.user: Optional[RawUser] = RawUser(data['user']) if data.get(
            'user'
        ) else None
        self.note: Optional[RawNote] = RawNote(data['note']) if data.get(
            'note'
        ) else None
        self.reaction: str = data['reaction']


class RawNote:
    """
    Attributes
    ----------
    id :  str
    created_at :  datetime
    user_id :  str
    author :  RawUser
    content : Optional[str]
    cw : Optional[str]
    renote : Optional[RawRenote]
    visibility : Optional[str]
    renote_count : Optional[int]
    replies_count : Optional[int]
    reactions : Optional[Dict[str, Any]]
    emojis : List[RawEmoji]
    file_ids : Optional[List[str]]
    files : List[RawFile]
    reply_id : Optional[str]
    renote_id : Optional[str]
    poll : Optional[RawPoll]
    visible_user_ids : Optional[List[str]]
    via_mobile :  bool
    local_only :  bool
    extract_mentions :  bool
    extract_hashtags :  bool
    extract_emojis :  bool
    preview :  bool
    media_ids : Optional[List[str]]
    field : Optional[dict]
    tags : Optional[List[str]]
    channel_id : Optional[str]
    """

    __slots__ = (
        'id',
        'created_at',
        'user_id',
        'author',
        'content',
        'cw',
        'renote',
        'visibility',
        'renote_count',
        'replies_count',
        'reactions',
        'emojis',
        'file_ids',
        'files',
        'reply_id',
        'renote_id',
        'uri',
        'poll',
        'visible_user_ids',
        'via_mobile',
        'local_only',
        'extract_mentions',
        'extract_hashtags',
        'extract_emojis',
        'preview',
        'media_ids',
        'field',
        'tags',
        'channel_id',
    )

    def __init__(self, data: NotePayload):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.user_id: str = data['user_id']
        self.author: RawUser = RawUser(data['user'])
        self.content: Optional[str] = data.get('text')
        self.cw: Optional[str] = data.get('cw')
        self.renote: Optional[RawRenote] = RawRenote(
            data['renote']
        ) if data.get('renote') else None
        self.visibility: Optional[str] = data.get(
            'visibility'
        )  # This may be an optional
        self.renote_count: Optional[int] = data.get(
            'renote_count'
        )  # TODO: Optionalかどうか
        self.replies_count: Optional[int] = data.get(
            'replies_count'
        )  # TODO: Optionalかどうか
        self.reactions: Dict[str, Any] = data['reactions']
        self.emojis: List[RawEmoji] = [RawEmoji(i) for i in data['emojis']]
        self.file_ids: Optional[List[str]] = data['file_ids']
        self.files: List[RawFile] = [
            RawFile(upper_to_lower(i)) for i in data['files']
        ]
        self.reply_id: Optional[str] = data['reply_id']
        self.renote_id: Optional[str] = data['renote_id']
        self.poll: Optional[RawPoll] = RawPoll(data['poll']) if data.get(
            'poll'
        ) else None
        self.visible_user_ids: Optional[List[str]] = data.get(
            'visible_user_ids', []
        )
        self.via_mobile: bool = bool(data.get('via_mobile', False))
        self.local_only: bool = bool(data.get('local_only', False))
        self.extract_mentions: bool = bool(data.get('extract_mentions'))
        self.extract_hashtags: bool = bool(data.get('extract_hashtags'))
        self.extract_emojis: bool = bool(data.get('extract_emojis'))
        self.preview: bool = bool(data.get('preview'))
        self.media_ids: Optional[List[str]] = data.get('media_ids')
        self.field: Optional[Dict[Any, Any]] = {}
        self.tags: Optional[List[str]] = data.get('tags', [])
        self.channel_id: Optional[str] = data.get('channel_id')



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac import AbstractManager
from mipac.actions.user import UserActions
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions
    from mipac.models.user import User


class UserManager(AbstractManager):
    def __init__(
        self,
        user: Optional[User] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.user: Optional[User] = user

    @property
    def action(self) -> UserActions:
        return UserActions(
            session=self.__session, client=self.__client, user=self.user
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac import AbstractManager
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.client import ClientActions


class AdminModeratorManager(AbstractManager):
    def __init__(
        self,
        user_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__user_id: Optional[str] = user_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add(self, user_id: Optional[str] = None) -> bool:
        """
        Add a user as a moderator

        Parameters
        ----------
        user_id : Optional[str], default=None
            ユーザーのID

        Returns
        -------
        bool
            成功したか否か
        """

        user_id = user_id or self.__user_id
        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/moderators/add'),
            json=data,
            auth=True,
            lower=True,
        )
        return bool(res)

    async def remove(self, user_id: Optional[str] = None) -> bool:
        """
        Unmoderate a user

        Parameters
        ----------
        user_id : Optional[str], default=None
            ユーザーのID

        Returns
        -------
        bool
            成功したか否か
        """
        user_id = user_id or self.__user_id
        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/moderators/remove'),
            json=data,
            auth=True,
            lower=True,
        )
        return bool(res)



from typing import Any, Dict, List, Optional, TypedDict

from .drive import FilePayload
from .emoji import EmojiPayload
from .user import UserPayload

__all__ = (
    'NotePayload',
    'GeoPayload',
    'ReactionPayload',
    'PollPayload',
    'RenotePayload',
    'OptionalReaction',
)


class GeoPayload(TypedDict):
    """
    衛星情報
    """

    coordinates: Optional[List[Any]]
    altitude: Optional[int]
    accuracy: Optional[int]
    altitude_accuracy: Optional[int]
    heading: Optional[int]
    speed: Optional[int]


class PollPayload(TypedDict, total=False):
    """
    アンケート情報
    """

    multiple: bool
    expires_at: int
    choices: List[str]
    expired_after: int


class RenotePayload(TypedDict):
    id: str
    created_at: str
    user_id: str
    user: UserPayload
    text: str
    cw: str
    visibility: str
    renote_count: int
    replies_count: Optional[int]
    reactions: Dict[str, Any]
    emojis: Optional[List]
    file_ids: Optional[List]
    files: Optional[List]
    reply_id: Optional[str]
    renote_id: Optional[str]
    uri: Optional[str]
    poll: Optional[PollPayload]
    tags: Optional[List[str]]
    channel_id: Optional[str]


class _NoteOptional(TypedDict, total=False):
    """
    ノートに必ず存在すると限らない物
    """

    text: str
    cw: str
    geo: GeoPayload


class NotePayload(_NoteOptional):
    """
    note object
    """

    id: str
    created_at: str
    user_id: str
    user: UserPayload
    visibility: Optional[str]
    renote_count: Optional[int]
    replies_count: Optional[int]
    reactions: Dict[str, Any]
    emojis: List[EmojiPayload]
    file_ids: Optional[List[str]]
    files: Optional[List[FilePayload]]
    reply_id: Optional[str]
    renote_id: Optional[str]
    poll: Optional[PollPayload]
    visible_user_ids: Optional[List[str]]
    via_mobile: Optional[bool]
    local_only: Optional[bool]
    extract_mentions: Optional[bool]
    extract_hashtags: Optional[bool]
    extract_emojis: Optional[bool]
    preview: Optional[bool]
    media_ids: Optional[List[str]]
    renote: Optional[RenotePayload]
    field: Optional[dict]
    tags: Optional[List[str]]
    channel_id: Optional[str]


class OptionalReaction(TypedDict, total=False):
    created_at: str
    type: str
    is_read: bool
    user: UserPayload
    note: NotePayload
    id: str


class ReactionPayload(OptionalReaction):
    reaction: str



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

__all__ = ['FollowManager', 'FollowRequestManager']

from mipac import AbstractManager, HTTPClient
from mipac.http import Route
from mipac.models.user import FollowRequest, User

if TYPE_CHECKING:
    from mipac.client import ClientActions


class FollowManager(AbstractManager):
    def __init__(
        self,
        user_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__user_id: Optional[str] = user_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add(
        self, user_id: Optional[str] = None
    ) -> tuple[bool, Optional[str]]:
        """
        ユーザーをフォローします

        Returns
        -------
        bool
            成功ならTrue, 失敗ならFalse
        str
            実行に失敗した際のエラーコード
        """

        user_id = user_id or self.__user_id

        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/following/create'),
            json=data,
            auth=True,
            lower=True,
        )
        if res.get('error'):
            code = res['error']['code']
            status = False
        else:
            code = None
            status = True
        return status, code

    async def remove(self, user_id: Optional[str] = None) -> bool:
        """
        ユーザーのフォローを解除します

        Returns
        -------
        bool
            成功ならTrue, 失敗ならFalse
        """

        user_id = user_id or self.__user_id

        data = {'userId': user_id}
        res = await self.__session.request(
            Route('POST', '/api/following/delete'), json=data, auth=True
        )
        return bool(res.status_code == 204 or 200)


class FollowRequestManager(AbstractManager):
    def __init__(
        self,
        user_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__user_id: Optional[str] = user_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def get_all(self) -> List[FollowRequest]:
        """
        未承認のフォローリクエストを取得します
        """

        return [
            FollowRequest(i['follower'])
            for i in await self.__session.request(
                Route('POST', '/api/following/requests/list'),
                auth=True,
                lower=True,
            )
        ]

    async def get_user(self, user_id: Optional[str] = None) -> User:
        """
        フォローリクエスト元のユーザーを取得します
        Parameters
        ----------
        user_id : Optional[str], default=None
            ユーザーID

        Returns
        -------
        User
            フォローリクエスト元のユーザー
        """

        user_id = user_id or self.__user_id

        return await self.__client.user.action.get(user_id)

    async def accept(self, user_id: Optional[str] = None) -> bool:
        """
        与えられたIDのユーザーのフォローリクエストを承認します
        """

        user_id = user_id or self.__user_id

        data = {'userId': user_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/following/requests/accept'),
                json=data,
                auth=True,
            )
        )

    async def reject(self, user_id: Optional[str]) -> bool:
        """
        与えられたIDのユーザーのフォローリクエストを拒否します
        """

        user_id = user_id or self.__user_id

        data = {'userId': user_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/following/requests/reject'),
                json=data,
                auth=True,
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncIterator, Dict, List, Optional

from mipac.core.models.emoji import RawEmoji
from mipac.core.models.instance import RawInstance
from mipac.models.emoji import Emoji
from mipac.types import MetaPayload

if TYPE_CHECKING:
    from mipac.client import ClientActions
    from mipac.models.user import User


__all__ = ('InstanceMeta', 'Instance')


class InstanceMeta:
    def __init__(self, data: MetaPayload):
        self.maintainer_name: str = data['maintainer_name']
        self.maintainer_email: str = data['maintainer_email']
        self.version: str = data['version']
        self.name: str = data['name']
        self.uri: str = data['uri']
        self.description: str = data['description']
        self.langs: List[str] = data['langs']
        self.tos_url: Optional[str] = data['tos_url']
        self.repository_url: str = data['repository_url']
        self.feedback_url: str = data['feedback_url']
        self.secure: bool = bool(data['secure'])
        self.disable_registration: bool = bool(data['disable_registration'])
        self.disable_local_timeline: bool = bool(
            data['disable_local_timeline']
        )
        self.disable_global_timeline: bool = bool(
            data['disable_global_timeline']
        )
        self.drive_capacity_per_local_user_mb: int = data[
            'drive_capacity_per_local_user_mb'
        ]
        self.drive_capacity_per_remote_user_mb: int = data[
            'drive_capacity_per_remote_user_mb'
        ]
        self.email_required_for_signup: bool = bool(
            data['email_required_for_signup']
        )
        self.enable_hcaptcha: bool = bool(data['enable_hcaptcha'])
        self.enable_recaptcha: bool = bool(data['enable_recaptcha'])
        self.recaptcha_site_key: str = data['recaptcha_site_key']
        self.sw_publickey: str = data['sw_publickey']
        self.mascot_image_url: str = data['mascot_image_url']
        self.error_image: str = data['error_image_url']
        self.max_note_text_length: int = data['max_note_text_length']
        self.emojis: List[Emoji] = [Emoji(RawEmoji(i)) for i in data['emojis']]
        self.ads: list = data['ads']
        self.enable_email: bool = bool(data['enable_email'])
        self.enable_twitter_integration = bool(
            data['enable_twitter_integration']
        )
        self.enable_github_integration: bool = bool(
            data['enable_github_integration']
        )
        self.enable_discord_integration: bool = bool(
            data['enable_discord_integration']
        )
        self.enable_service_worker: bool = bool(data['enable_service_worker'])
        self.translator_available: bool = bool(data['translator_available'])
        self.pinned_page: Optional[List[str]] = data.get('pinned_page')
        self.cache_remote_files: Optional[bool] = data.get(
            'cache_remote_files'
        )
        self.proxy_remote_files: Optional[bool] = data.get(
            'proxy_remote_files'
        )
        self.require_setup: Optional[bool] = data.get('require_setup')
        self.features: Optional[Dict[str, bool]] = data.get('features')


class Instance:
    def __init__(self, raw_data: RawInstance, *, client: ClientActions):
        """
        インスタンス情報
        
        Parameters
        ----------
        raw_data : RawInstance
            インスタンス情報の入った dict
        """

        self.__raw_data: RawInstance = raw_data
        self.__client: ClientActions = client

    @property
    def host(self):
        return self.__raw_data.host

    @property
    def name(self):
        return self.__raw_data.name

    @property
    def software_name(self):
        return self.__raw_data.software_name

    @property
    def software_version(self):
        return self.__raw_data.software_version

    @property
    def icon_url(self):
        return self.__raw_data.icon_url

    @property
    def favicon_url(self):
        return self.__raw_data.favicon_url

    @property
    def theme_color(self):
        return self.__raw_data.theme_color

    def get_users(
        self,
        limit: int = 10,
        *,
        offset: int = 0,
        sort: Optional[str] = None,
        state: str = 'all',
        origin: str = 'local',
        username: Optional[str] = None,
        hostname: Optional[str] = None,
        get_all: bool = False
    ) -> AsyncIterator[User]:
        """

        Parameters
        ----------
        limit: int
        offset:int
        sort:str
        state:str
        origin:str
        username:str
        hostname:str
        get_all:bool

        Returns
        -------
        AsyncIterator[User]
        """
        # return self.__client.get_users(limit=limit, offset=offset, sort=sort, state=state, origin=origin, username=username,
        #                                hostname=hostname, get_all=get_all)  # TODO: 修正



from __future__ import annotations

from datetime import datetime
from typing import List

from mipac.types import ChatPayload

from .user import RawUser

__all__ = ('RawChat',)


class RawChat:
    """
    Attributes
    ----------
    id : str
        chat ID
    created_at : str
        chat creation time
    content : str
        chat content
    user_id : str
       ID of the user who created the chat
    author : RawUser
        user who created the chat
    recipient_id : str
    recipient : str
    group_id : str
        group ID
    file_id : str
        ID of the attached file
    is_read : bool
        whether the chat is read
    reads : List
    """

    __slots__ = (
        'id',
        'created_at',
        'content',
        'user_id',
        'author',
        'recipient_id',
        'recipient',
        'group_id',
        'file_id',
        'is_read',
        'reads',
    )

    def __init__(self, data: ChatPayload):
        """

        Parameters
        ----------
        data: ChatPayload
        """
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.content: str = data['text']
        self.user_id: str = data['user_id']
        self.author: RawUser = RawUser(data['user'])
        self.recipient_id: str = data['recipient_id']
        self.recipient: str = data['recipient']
        self.group_id: str = data['group_id']
        self.file_id: str = data['file_id']
        self.is_read: bool = bool(data['is_read'])
        self.reads: List = data['reads']



from typing import List, Optional

from mipac.core.models.drive import RawFile
from mipac.http import Route

# __all__ = ('MiFile', 'check_upload', 'get_file_ids')


class MiFile:
    __slots__ = (
        'path',
        'file_id',
        'name',
        'folder_id',
        'comment',
        'is_sensitive',
        'force',
    )

    def __init__(
        self,
        path: Optional[str] = None,
        file_id: Optional[str] = None,
        name: Optional[str] = None,
        folder_id: Optional[str] = None,
        comment: Optional[str] = None,
        is_sensitive: bool = False,
        force: bool = False,
    ):
        """
        Parameters
        ----------
        path : Optional[str], default=None
            path to a local file
        file_id : Optional[str], default=None
            ID of the file that exists on the drive
        name Optional[str], default=None
            file name
        folder_id : Optional[str], default=None
            Folder ID
        comment : Optional[str], default=None
            Comments on files
        is_sensitive : Optional[str], default=None
            Whether this item is sensitive
        force : bool, default=False
            Whether to force overwriting even if it already exists on the drive
        """
        self.path = path
        self.file_id = file_id
        self.name = name
        self.folder_id = folder_id
        self.comment = comment
        self.is_sensitive = is_sensitive
        self.force = force


# async def check_upload(files: List[MiFile]):
#     _files = []
#     for file in files:
#         if file.path:
#             endpoint = Route('POST', '/api/drive/files/create')
#             data = {'file': open(file.path, 'rb'),
#                     'name': file.name,
#                     'folderId': file.folder_id,
#                     'isSensitive': file.is_sensitive,
#                     'comment': file.comment,
#                     'force': file.force}
#             _files.append(RawFile(await HTTPSession.request(endpoint, auth=True, data=data, lower=True)).id)  # TODO: 治す
#         else:
#             _files.append(file.file_id)
#
#     return _files


# async def get_file_ids(files: List[MiFile]):
#     return await check_upload(files)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from mipac.core.models.note import RawNote, RawReaction, RawRenote
from mipac.core.models.poll import RawPoll
from mipac.core.models.reaction import RawNoteReaction
from mipac.core.models.user import RawUser
from mipac.exception import NotExistRequiredData
from mipac.models.user import User

if TYPE_CHECKING:
    from mipac.actions.note import NoteActions
    from mipac.manager.client import ClientActions
    from mipac.models.drive import File
    from mipac.models.emoji import Emoji

__all__ = (
    'Note',
    'Poll',
    'Reaction',
    'Follow',
    'Header',
    'Renote',
    'NoteReaction',
)


class Follow:
    def __init__(self, data):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.type: Optional[str] = data.get('type')
        self.user: Optional[User] = data.get('user')

    async def follow(self) -> tuple[bool, Optional[str]]:
        """
        ユーザーをフォローします
        Returns
        -------
        bool
            成功ならTrue, 失敗ならFalse
        str
            実行に失敗した際のエラーコード
        """

        if self.id:
            raise NotExistRequiredData('user_idがありません')
        return await self._state.user.follow.add(user_id=self.id)

    async def unfollow(self, user_id: Optional[str] = None) -> bool:
        """
        与えられたIDのユーザーのフォローを解除します

        Parameters
        ----------
        user_id : Optional[str] = None
            フォローを解除したいユーザーのID

        Returns
        -------
        status
            成功ならTrue, 失敗ならFalse
        """

        if user_id is None:
            user_id = self.user.id
        return await self._state.user.follow.remove(user_id)


class Header:
    def __init__(self, data):
        self.id = data.get('id')
        self.type = data.get('type')


class Poll:
    def __init__(self, raw_data: RawPoll):
        self.__raw_data = raw_data

    @property
    def multiple(self) -> bool:
        return self.__raw_data.multiple

    @property
    def expires_at(self) -> Optional[int]:
        return self.__raw_data.expires_at

    @property
    def choices(self):
        return self.__raw_data.choices

    @property
    def expired_after(self) -> Optional[int]:
        return self.__raw_data.expired_after


class Renote:
    def __init__(self, raw_data: RawRenote):
        self.__raw_data: RawRenote = raw_data
        self.__client = manager.ClientActions()

    @property
    def id(self) -> str:
        return self.__raw_data.id

    @property
    def created_at(self) -> datetime:
        return self.__raw_data.created_at

    @property
    def user_id(self) -> str:
        return self.__raw_data.user_id

    @property
    def user(self) -> User:
        return User(self.__raw_data.user, client=self.__client)

    @property
    def content(self) -> Optional[str]:
        return self.__raw_data.content

    @property
    def cw(self) -> Optional[str]:
        return self.__raw_data.cw

    @property
    def visibility(self) -> str:
        return self.__raw_data.visibility

    @property
    def renote_count(self) -> int:
        return self.__raw_data.renote_count

    @property
    def replies_count(self) -> int:
        return self.__raw_data.replies_count

    @property
    def reactions(self):
        return self.__raw_data.reactions

    @property
    def emojis(self):
        return self.__raw_data.emojis

    @property
    def file_ids(self):
        return self.__raw_data.file_ids

    @property
    def files(self):
        return self.__raw_data.files

    @property
    def reply_id(self) -> Optional[str]:
        return self.__raw_data.reply_id

    @property
    def renote_id(self) -> Optional[str]:
        return self.__raw_data.renote_id

    @property
    def uri(self) -> Optional[str]:
        return self.__raw_data.uri

    @property
    def poll(self) -> Union[Poll, None]:
        return Poll(self.__raw_data.poll) if self.__raw_data.poll else None

    async def delete(self) -> bool:
        return await self.__client.note.delete(self.__raw_data.id)


class NoteReaction:
    def __init__(self, raw_data: RawNoteReaction):
        self.__raw_data = raw_data

    @property
    def id(self) -> str:
        return self.__raw_data.id

    @property
    def created_at(self) -> datetime:
        return self.__raw_data.created_at

    @property
    def user(self) -> User:
        return User(RawUser(self.__raw_data.user), client=self.__client)

    @property
    def reaction(self) -> str:
        return self.__raw_data.reaction


class Reaction:
    def __init__(self, raw_data: RawReaction, *, client: ClientActions):
        self.__raw_data: RawReaction = raw_data
        self.__client: ClientActions = client

    @property
    def id(self) -> Optional[str]:
        return self.__raw_data.id

    @property
    def created_at(self) -> Optional[datetime]:
        return self.__raw_data.created_at

    @property
    def type(self) -> Optional[str]:
        return self.__raw_data.type

    @property
    def is_read(self) -> bool:
        return self.__raw_data.is_read

    @property
    def user(self) -> Optional[User]:
        return (
            User(self.__raw_data.user, client=self.__client)
            if self.__raw_data.user
            else None
        )

    @property
    def note(self) -> Optional[Note]:
        return (
            Note(self.__raw_data.note, client=self.__client)
            if self.__raw_data.note
            else None
        )

    @property
    def reaction(self) -> str:
        return self.__raw_data.reaction

    # @property  # TODO: 修正
    # def action(self) -> ReactionManager:
    #     return manager.ClientActions().reaction


class Note:
    def __init__(self, raw_data: RawNote, client: ClientActions):
        self.__raw_data: RawNote = raw_data
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        """
        ユーザーのID

        Returns
        -------
        str
            ユーザーのID
        """
        return self.__raw_data.id

    @property
    def created_at(self) -> datetime:
        return self.__raw_data.created_at

    @property
    def user_id(self) -> str:
        return self.__raw_data.user_id

    @property
    def author(self) -> User:
        return User(self.__raw_data.author, client=self.__client)

    @property
    def content(self) -> Optional[str]:
        return self.__raw_data.content

    @property
    def cw(self) -> Optional[str]:
        return self.__raw_data.cw

    @property
    def renote(self) -> Union[None, Renote]:
        return (
            Renote(self.__raw_data.renote) if self.__raw_data.renote else None
        )

    @property
    def visibility(self) -> Optional[str]:
        return self.__raw_data.visibility

    @property
    def renote_count(self) -> Optional[int]:
        return self.__raw_data.renote_count

    @property
    def replies_count(self) -> Optional[int]:
        return self.__raw_data.replies_count

    @property
    def reactions(self) -> Optional[Dict[str, Any]]:  # TODO: 型の確認
        return self.__raw_data.reactions

    @property
    def emojis(self) -> List[Emoji]:
        return [Emoji(i) for i in self.__raw_data.emojis]

    @property
    def file_ids(self) -> Optional[List[str]]:
        return self.__raw_data.file_ids

    @property
    def files(self) -> List[File]:
        return [File(i, client=self.__client) for i in self.__raw_data.files]

    @property
    def reply_id(self) -> Optional[str]:
        return self.__raw_data.reply_id

    @property
    def renote_id(self) -> Optional[str]:
        return self.__raw_data.renote_id

    @property
    def poll(self) -> Union[Poll, None]:
        return Poll(self.__raw_data.poll) if self.__raw_data.poll else None

    @property
    def visible_user_ids(self) -> Optional[List[str]]:
        return self.__raw_data.visible_user_ids

    @property
    def via_mobile(self) -> bool:
        return self.__raw_data.via_mobile

    @property
    def local_only(self) -> bool:
        return self.__raw_data.local_only

    @property
    def extract_mentions(self) -> bool:
        return self.__raw_data.extract_mentions

    @property
    def extract_hashtags(self) -> bool:
        return self.__raw_data.extract_hashtags

    @property
    def extract_emojis(self) -> bool:
        return self.__raw_data.extract_emojis

    @property
    def preview(self) -> bool:
        return self.__raw_data.preview

    @property
    def media_ids(self) -> Optional[List[str]]:
        return self.__raw_data.media_ids

    @property
    def field(self) -> Optional[Dict[Any, Any]]:  # TODO: any
        return self.__raw_data.field

    @property
    def tags(self) -> Optional[List[str]]:
        return self.__raw_data.tags

    @property
    def channel_id(self) -> Optional[str]:
        return self.__raw_data.channel_id

    @property
    def action(self) -> NoteActions:
        """
        ノートに対するアクション

        Returns
        -------
        NoteActions
        """
        return self.__client._create_note_instance(self.id).action

    async def reply(
        self,
        content: Optional[str],
        cw: Optional[str] = None,
        extract_mentions: bool = True,
        extract_hashtags: bool = True,
        extract_emojis: bool = True,
        renote_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        file_ids=None,
        poll: Optional[Poll] = None,
    ) -> Note:
        """
        ノートに対して返信を送信します

        Parameters
        ----------
        content: Optional[str]
            返信内容
        cw: Optional[str]
            閲覧注意
        extract_mentions : bool, optional
            メンションを展開するか, by default False
        extract_hashtags : bool, optional
            ハッシュタグを展開するか, by default False
        extract_emojis : bool, optional
            絵文字を展開するか, by default False
        renote_id : Optional[str], optional
            リノート先のid, by default None
        channel_id : Optional[str], optional
            チャンネルid, by default None
        file_ids : [type], optional
            添付するファイルのid, by default None
        poll : Optional[Poll], optional
            アンケート, by default None
        """
        if file_ids is None:
            file_ids = []
        visibility = self.visibility or 'public'

        return await self.__client.note.action.send(
            content,
            visibility=visibility,
            visible_user_ids=self.visible_user_ids,
            cw=cw,
            local_only=self.local_only,
            extract_mentions=extract_mentions,
            extract_hashtags=extract_hashtags,
            extract_emojis=extract_emojis,
            reply_id=self.id,
            renote_id=renote_id,
            channel_id=channel_id,
            # file_ids=file_ids, TODO:  修正
            poll=poll,
        )



from __future__ import annotations

import json
import sys
from typing import Any, Dict

import aiohttp

from mipac import __version__
from mipac.exception import APIError
from mipac.util import remove_dict_empty, upper_to_lower


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()


async def json_or_text(response: aiohttp.ClientResponse):
    text = await response.text(encoding='utf-8')
    try:
        if 'application/json' in response.headers['Content-Type']:
            return json.loads(text)
    except KeyError:
        pass


class Route:
    def __init__(self, method: str, path: str):
        self.path: str = path
        self.method: str = method
        # self.url = config.i.origin_uri + path


class HTTPClient:
    def __init__(self, url: str, token: str) -> None:
        user_agent = 'Misskey Bot (https://github.com/yupix/MiPA {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self.user_agent = user_agent.format(
            __version__, sys.version_info, aiohttp.__version__
        )
        self.__session: aiohttp.ClientSession = MISSING
        self.__url: str = url
        self.__token: str = token

    @property
    def session(self) -> aiohttp.ClientSession:
        return self.__session

    async def request(self, route: Route, **kwargs) -> Any:
        headers: Dict[str, str] = {
            'User-Agent': self.user_agent,
        }

        is_lower = kwargs.pop('lower') if kwargs.get('lower') else False

        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['json'] = kwargs.pop('json')

        if kwargs.get('auth') and kwargs.pop('auth'):
            key = (
                'json' if 'json' in kwargs or 'data' not in kwargs else 'data'
            )
            if not kwargs.get(key):
                kwargs[key] = {}
            kwargs[key]['i'] = self.__token

        for i in ('json', 'data'):
            if kwargs.get(i):
                kwargs[i] = remove_dict_empty(kwargs[i])

        async with self.__session.request(
            route.method, self.__url + route.path, **kwargs
        ) as res:
            data = await json_or_text(res)
            if is_lower:
                if isinstance(data, list):
                    data = [upper_to_lower(i) for i in data]
                else:
                    data = upper_to_lower(data)
            if 300 > res.status >= 200:
                return data
            if 511 > res.status >= 300:
                raise APIError(data)

    async def close_session(self):
        await self.__session.close()

    async def login(self):
        self.__session = aiohttp.ClientSession()
        data = await self.request(Route('POST', '/api/i'), auth=True)
        return data



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from mipac import AbstractManager
from mipac.exception import NotExistRequiredData
from mipac.http import HTTPClient, Route
from mipac.util import check_multi_arg

if TYPE_CHECKING:
    from mipac.client import ClientActions


class AdminEmojiManager(AbstractManager):
    def __init__(
        self,
        emoji_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.emoji_id: Optional[str] = emoji_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add(
        self,
        file_id: Optional[str] = None,
        *,
        name: Optional[str] = None,
        url: Optional[str] = None,
        category: Optional[str] = None,
        aliases: Optional[List[str]] = None
    ) -> bool:
        if config.is_ayuskey:  # TODO: どうにかする
            data = {
                'name': name,
                'url': url,
                'category': category,
                'aliases': aliases,
            }
        else:
            data = {'fileId': file_id}

        if not check_multi_arg(file_id, url):
            raise NotExistRequiredData('required a file_id or url')
        return bool(
            await self.__session.request(
                Route('POST', '/api/admin/emoji/add'),
                json=data,
                lower=True,
                auth=True,
            )
        )

    async def remove(self, emoji_id: Optional[str] = None) -> bool:
        emoji_id = emoji_id or self.emoji_id

        if emoji_id is None:
            raise NotExistRequiredData('idが不足しています')

        return bool(
            await self.__session.request(
                Route('POST', '/api/admin/emoji/remove'),
                json={'id': emoji_id},
                lower=True,
                auth=True,
            )
        )



from __future__ import annotations

from typing import Optional

from mipac.types import InstancePayload

__all__ = ['RawInstance']


class RawInstance:
    """
    Attributes
    ----------
    host : Optional[str], default=None
    name : Optional[str], default=None
    software_name : Optional[str], default=None
    software_version : Optional[str], default=None
    icon_url : Optional[str], default=None
    favicon_url : Optional[str], default=None
    theme_color : Optional[str], default=None
    """

    __slots__ = (
        'host',
        'name',
        'software_name',
        'software_version',
        'icon_url',
        'favicon_url',
        'theme_color',
    )

    def __init__(self, data: InstancePayload):
        self.host: Optional[str] = data.get('host')
        self.name: Optional[str] = data.get('name')
        self.software_name: Optional[str] = data.get('software_version')
        self.software_version: Optional[str] = data.get('software_version')
        self.icon_url: Optional[str] = data.get('icon_url')
        self.favicon_url: Optional[str] = data.get('favicon_url')
        self.theme_color: Optional[str] = data.get('theme_color')



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from mipac.core.models.user import RawUser
from mipac.models.emoji import Emoji
from mipac.models.instance import Instance
from mipac.types import (ChannelPayload, FieldContentPayload,
                         PinnedNotePayload, PinnedPagePayload)

if TYPE_CHECKING:
    from mipac.actions.user import UserActions
    from mipac.manager.client import ClientActions


__all__ = ['User', 'FollowRequest', 'Followee']


class Followee:
    def __init__(self, data, *, client: ClientActions):
        self.id: str = data['id']
        self.created_at: datetime = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.followee_id: str = data['followee_id']
        self.follower_id: str = data['follower_id']
        self.user: User = User(RawUser(data['follower']), client=client)


class FollowRequest:
    def __init__(self, data):
        self.id = data['id']
        self.name = data['username']
        self.nickname = data['name']
        self.host = data['host']
        self.avatar_url = data['avatar_url']
        self.avatar_blurhash = data['avatar_blurhash']
        self.avatar_color = data['avatar_color']
        self.emojis = data['emojis']
        self.online_status = data['online_status']
        self.is_admin: bool = bool(data.get('is_admin'))
        self.is_bot: bool = bool(data.get('is_bot'))
        self.is_cat: bool = bool(data.get('is_cat'))
        # self.__client = manager.ClientActions()

    # @property
    # def action(self) -> FollowRequestManager:
    #     return self.__client.user.get_follow_request(self.id)

    # async def get_profile(self) -> User:
    #     return await self.__client.user.get(user_id=self.id)


class Channel:
    def __init__(self, data: ChannelPayload):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.last_noted_at: Optional[str] = data.get('last_noted_at')
        self.name: Optional[str] = data.get('name')
        self.description: Optional[str] = data.get('description')
        self.banner_url: Optional[str] = data.get('banner_url')
        self.notes_count: Optional[int] = data.get('notes_count')
        self.users_count: Optional[int] = data.get('users_count')
        self.is_following: Optional[bool] = data.get('is_following')
        self.user_id: Optional[str] = data.get('user_id')


class PinnedNote:
    def __init__(self, data: PinnedNotePayload, *, client: ClientActions):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.text: Optional[str] = data.get('text')
        self.cw: Optional[str] = data.get('cw')
        self.user_id: Optional[str] = data.get('user_id')
        self.user: Optional[User] = User(
            RawUser(data['user']), client=client
        ) if data.get('user') else None
        self.reply_id: Optional[str] = data.get('reply_id')
        self.reply: Optional[Dict[str, Any]] = data.get('reply')
        self.renote: Optional[Dict[str, Any]] = data.get('renote')
        self.via_mobile: Optional[bool] = data.get('via_mobile')
        self.is_hidden: Optional[bool] = data.get('is_hidden')
        self.visibility: Optional[bool] = bool(data['visibility']) if data.get(
            'visibility'
        ) else None
        self.mentions: Optional[List[str]] = data.get('mentions')
        self.visible_user_ids: Optional[List[str]] = data.get(
            'visible_user_ids'
        )
        self.file_ids: Optional[List[str]] = data.get('file_ids')
        self.files: Optional[List[str]] = data.get('files')
        self.tags: Optional[List[str]] = data.get('tags')
        self.poll: Optional[List[str]] = data.get('poll')
        self.channel: Optional[Channel] = Channel(data['channel']) if data.get(
            'channel'
        ) else None
        self.local_only: Optional[bool] = data.get('local_only')
        self.emojis: Optional[List[Emoji]] = [
            Emoji(i) for i in data['emojis']
        ] if data.get('emojis') else None
        self.reactions: Optional[Dict[str, Any]] = data.get('reactions')
        self.renote_count: Optional[int] = data.get('renote_count')
        self.replies_count: Optional[int] = data.get('replies_count')
        self.uri: Optional[str] = data.get('uri')
        self.url: Optional[str] = data.get('url')
        self.my_reaction: Optional[Dict[str, Any]] = data.get('my_reaction')


class PinnedPage:
    def __init__(self, data: PinnedPagePayload):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.updated_at: Optional[str] = data.get('updated_at', None)
        self.title: Optional[str] = data.get('title')
        self.name: Optional[str] = data.get('name')
        self.summary: Optional[str] = data.get('summary')
        self.content: Optional[List] = data.get('content')
        self.variables: Optional[List] = data.get('variables')
        self.user_id: Optional[str] = data.get('user_id')
        self.author: Optional[Dict[str, Any]] = data.get('author')


class FieldContent:
    def __init__(self, data: FieldContentPayload):
        self.name: str = data['name']
        self.value: str = data['value']


class User:
    def __init__(self, raw_user: RawUser, *, client: ClientActions):
        self.__raw_user = raw_user
        self.__client: ClientActions = client

    @property
    def id(self):
        return self.__raw_user.id

    @property
    def name(self):
        return self.__raw_user.name

    @property
    def nickname(self):
        return self.__raw_user.nickname

    @property
    def host(self):
        return self.__raw_user.host

    @property
    def avatar_url(self):
        return self.__raw_user.avatar_url

    @property
    def is_admin(self):
        return self.__raw_user.is_admin

    @property
    def is_moderator(self):
        return self.__raw_user.is_moderator

    @property
    def is_bot(self):
        return self.__raw_user.is_bot

    @property
    def is_cat(self):
        return self.__raw_user.is_cat

    @property
    def is_lady(self):
        return self.__raw_user.is_lady

    @property
    def emojis(self):
        return self.__raw_user.emojis

    @property
    def online_status(self):
        return self.__raw_user.online_status

    @property
    def url(self):
        return self.__raw_user.url

    @property
    def uri(self):
        return self.__raw_user.uri

    @property
    def created_at(self) -> datetime:
        return self.__raw_user.created_at

    @property
    def updated_at(self):
        return self.__raw_user.updated_at

    @property
    def is_locked(self):
        return self.__raw_user.is_locked

    @property
    def is_silenced(self):
        return self.__raw_user.is_silenced

    @property
    def is_suspended(self):
        return self.__raw_user.is_suspended

    @property
    def description(self):
        return self.__raw_user.description

    @property
    def location(self):
        return self.__raw_user.location

    @property
    def birthday(self):
        return self.__raw_user.birthday

    @property
    def fields(self):
        return self.__raw_user.fields

    @property
    def followers_count(self):
        return self.__raw_user.followers_count

    @property
    def following_count(self):
        return self.__raw_user.following_count

    @property
    def notes_count(self):
        return self.__raw_user.notes_count

    @property
    def pinned_note_ids(self):
        return self.__raw_user.pinned_note_ids

    @property
    def pinned_notes(self):
        return self.__raw_user.pinned_notes

    @property
    def pinned_page_id(self):
        return self.__raw_user.pinned_page_id

    @property
    def pinned_page(self):
        return self.__raw_user.pinned_page

    @property
    def ff_visibility(self):
        return self.__raw_user.ff_visibility

    @property
    def is_following(self):
        return self.__raw_user.is_following

    @property
    def is_follow(self):
        return self.__raw_user.is_follow

    @property
    def is_blocking(self):
        return self.__raw_user.is_blocking

    @property
    def is_blocked(self):
        return self.__raw_user.is_blocked

    @property
    def is_muted(self):
        return self.__raw_user.is_muted

    @property
    def details(self):
        return self.__raw_user.details

    @property
    def instance(self) -> Union[Instance, None]:
        return (
            Instance(self.__raw_user.instance, client=self.__client)
            if self.__raw_user.instance
            else None
        )

    @property
    def action(self) -> UserActions:
        return self.__client._create_user_instance(self).action



import pathlib

from setuptools import setup

description = "A Python wrapper for the Misskey API"
readme_file = pathlib.Path(__file__).parent / "README.md"
with readme_file.open(encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()





extras_require = {
    'dev': [
        'axblack',
        'isort',
        'mypy',
        'flake8'
    ],
    'ci': [
        'flake8',
        'mypy'
    ]
}

packages = [
    'mipac',
    'mipac.abc',
    'mipac.actions',
    'mipac.core',
    'mipac.core.models',
    'mipac.manager',
    'mipac.manager.admin',
    'mipac.models',
    'mipac.types'

]

setup(
    name="mipac",
    version="0.0.2",
    install_requires=requirements,
    url="https://github.com/yupix/mipac",
    author="yupix",
    author_email="yupi0982@outlook.jp",
    license="MIT",
    python_requires=">=3.10, <4.0",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=packages,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.10",
        "Natural Language :: Japanese",
        "License :: OSI Approved :: MIT License",
    ],
    extras_require=extras_require,
)



from __future__ import annotations

from typing import List, Optional

from mipac.types import EmojiPayload

__all__ = ('RawEmoji',)


class RawEmoji:
    """
    Attributes
    ----------
    id : Optional[str]
        絵文字のID
    aliases : Optional[List[str]]
        絵文字のエイリアス
    name : Optional[str]
        絵文字の名前
    category : Optional[str]
        絵文字のカテゴリ
    host : Optional[str]
        絵文字のホスト
    url : Optional[str]
        絵文字のURL
    """

    __slots__ = ('id', 'aliases', 'name', 'category', 'host', 'url')

    def __init__(self, data: EmojiPayload):
        self.id: Optional[str] = data.get('id')
        self.aliases: Optional[List[str]] = data.get('aliases')
        self.name: Optional[str] = data.get('name')
        self.category: Optional[str] = data.get('category')
        self.host: Optional[str] = data.get('host')
        self.url: Optional[str] = data.get('url')



from mipac.http import HTTPClient
from mipac.manager.client import ClientActions


class Client:
    def __init__(self, url: str, token: str):
        self.__url: str = url
        self.__token: str = token
        self.http: HTTPClient = HTTPClient(url, token)

    @property
    def action(self):
        return ClientActions(self.http)

    async def close_session(self):
        await self.http.close_session()



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from aiocache import Cache, cached

from mipac.core.models.note import RawNote
from mipac.core.models.user import RawUser
from mipac.exception import NotExistRequiredData, ParameterError
from mipac.http import HTTPClient, Route
from mipac.manager.note import NoteManager
from mipac.models.note import Note
from mipac.models.user import User
from mipac.util import (check_multi_arg, get_cache_key, key_builder,
                        remove_dict_empty)

__all__ = ['UserActions']

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions


class UserActions:
    def __init__(
        self,
        session: HTTPClient,
        client: ClientActions,
        user: Optional[User] = None,
    ):
        self.__session: HTTPClient = session
        self.__user: Optional[User] = user
        self.__client: ClientActions = client
        self.note: NoteManager = NoteManager(session=session, client=client)

    async def get_me(self) -> User:
        """
        ログインしているユーザーの情報を取得します
        """

        res = await self.__session.request(Route('POST', '/api/i'), auth=True)
        return User(RawUser(res), client=self.__client)

    @cached(ttl=10, namespace='get_user', key_builder=key_builder)
    async def get(
        self,
        user_id: Optional[str] = None,
        username: Optional[str] = None,
        host: Optional[str] = None,
    ) -> User:
        """
        ユーザーのプロフィールを取得します。一度のみサーバーにアクセスしキャッシュをその後は使います。
        fetch_userを使った場合はキャッシュが廃棄され再度サーバーにアクセスします。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        User
            ユーザー情報
        """

        field = remove_dict_empty(
            {'userId': user_id, 'username': username, 'host': host}
        )
        data = await self.__session.request(
            Route('POST', '/api/users/show'), json=field, auth=True, lower=True
        )
        return User(RawUser(data), client=self.__client)

    @get_cache_key
    async def fetch(
        self,
        user_id: Optional[str] = None,
        username: Optional[str] = None,
        host: Optional[str] = None,
        **kwargs,
    ) -> User:
        """
        サーバーにアクセスし、ユーザーのプロフィールを取得します。基本的には get_userをお使いください。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        User
            ユーザー情報
        """
        if not check_multi_arg(user_id, username):
            raise ParameterError('user_id, usernameどちらかは必須です')

        field = remove_dict_empty(
            {'userId': user_id, 'username': username, 'host': host}
        )
        data = await self.__session.request(
            Route('POST', '/api/users/show'), json=field, auth=True, lower=True
        )
        old_cache = Cache(namespace='get_user')
        await old_cache.delete(kwargs['cache_key'].format('get_user'))
        return User(RawUser(data), client=self.__client)

    async def get_notes(
        self,
        user_id: Optional[str] = None,
        include_replies: bool = True,
        limit: int = 10,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        since_date: int = 0,
        until_date: int = 0,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: Optional[List[str]] = None,
        exclude_nsfw: bool = True,
    ) -> List[Note]:

        if check_multi_arg(user_id, self.__user):
            raise ParameterError('user_idがありません')

        user_id = user_id or self.__user and self.__user.id
        data = {
            'userId': user_id,
            'includeReplies': include_replies,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'sinceDate': since_date,
            'untilDate': until_date,
            'includeMyRenotes': include_my_renotes,
            'withFiles': with_files,
            'fileType': file_type,
            'excludeNsfw': exclude_nsfw,
        }
        res = await self.__session.request(
            Route('POST', '/api/users/notes'), json=data, auth=True, lower=True
        )
        return [Note(RawNote(i), client=self.__client) for i in res]

    def get_mention(self, user: Optional[User] = None) -> str:
        """
        Get mention name of user.

        Parameters
        ----------
        user : Optional[User], default=None
            メンションを取得したいユーザーのオブジェクト

        Returns
        -------
        str
            メンション
        """

        user = user or self.__user

        if user is None:
            raise NotExistRequiredData('Required parameters: user')
        return (
            f'@{user.name}@{user.host}' if user.instance else f'@{user.name}'
        )

    # def get_follow(self, user_id: str) -> FollowManager:
    #     return FollowManager(user_id=user_id)
    #
    # def get_follow_request(self, user_id: str) -> FollowRequestManager:
    #     return FollowRequestManager(user_id=user_id)



from __future__ import annotations

from typing import TYPE_CHECKING, List, TypedDict

if TYPE_CHECKING:
    from mipac.types import UserPayload

__all__ = ('ChatPayload',)


class ChatPayload(TypedDict):
    id: str
    created_at: str
    text: str
    user_id: str
    user: UserPayload
    recipient_id: str
    recipient: str
    group_id: str
    file_id: str
    is_read: bool
    reads: List



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac import AbstractManager, HTTPClient
from mipac.http import Route

if TYPE_CHECKING:
    from mipac.client import ClientActions


class FavoriteManager(AbstractManager):
    def __init__(
        self,
        note_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__note_id = note_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add(self, note_id: Optional[str] = None) -> bool:
        note_id = note_id or self.__note_id
        data = {'noteId': note_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/notes/favorites/create'),
                json=data,
                auth=True,
            )
        )

    async def remove(self, note_id: Optional[str] = None) -> bool:
        note_id = note_id or self.__note_id
        data = {'noteId': note_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/notes/favorites/delete'),
                json=data,
                auth=True,
            )
        )



from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from mipac.core.models.instance import RawInstance
from mipac.types import UserPayload

__all__ = ('RawUserDetails', 'RawUser')


class RawUserDetails:
    """
    ユーザー情報だが、一般的に使うか怪しいもの

    Attributes
    ----------
    avatar_blurhash: Optional[str]
        ユーザーのアバターのblurhash
    avatar_color: str
        ユーザーのアバターの色
    lang: str
        ユーザーの言語
    """

    __slots__ = (
        'avatar_blurhash',
        'avatar_color',
        'banner_url',
        'banner_blurhash',
        'banner_color',
        'two_factor_enabled',
        'use_password_less_login',
        'security_keys',
        'has_pending_follow_request_from_you',
        'has_pending_follow_request_to_you',
        'public_reactions',
        'lang',
    )

    def __init__(self, data):
        self.avatar_blurhash: Optional[str] = data.get('avatar_blurhash')
        self.avatar_color: Optional[str] = data.get('avatar_color')
        self.banner_url = data.get('banner_url')
        self.banner_blurhash = data.get('banner_blurhash')
        self.banner_color = data.get('banner_color')
        self.two_factor_enabled = data.get('two_factor_enabled', False)
        self.use_password_less_login = data.get(
            'use_password_less_login', False
        )
        self.security_keys = data.get('security_keys', False)
        self.has_pending_follow_request_from_you = data.get(
            'has_pending_follow_request_from_you', False
        )
        self.has_pending_follow_request_to_you = data.get(
            'has_pending_follow_request_to_you', False
        )
        self.public_reactions = data.get('public_reactions', False)
        self.lang = data.get('lang')


class RawUser:
    """
    id : str
        ユーザーのID
    name : str
        ユーザーの名前
    nickname : Optional[str]
        ユーザーの表示名
    host : Optional[str]
        # TODO: いい感じに
    avatar_url : Optional[str]
        ユーザーのアバターurl
    is_admin : bool
        管理者か否か
    is_bot : bool
        ボットか否か
    is_cat : bool
        ねこか否か
    is_lady : bool
        お嬢様か否か (Ayuskeyのみ)
    emojis : Optional[List[str]]
        # TODO 謎
    url : Optional[str]
        # TODO 謎
    uri : Optional[str]
        # TODO 謎
    created_at : Optional[datetime]
        ユーザーの作成日時
    ff_visibility : str
        # TODO 謎
    is_following : bool
        フォローされてるか否か
    is_follow : bool
        フォローしているか否か
    is_blocking : bool
        ブロックしているか否か
    is_blocked : bool
        ブロックされてるか否か
    is_muted : bool
        ミュートしているか否か
    details : RawUserDetails
        ユーザーの詳細情報
    instance : Optional[RawInstance]
        インスタンスの情報
    """

    __slots__ = (
        'id',
        'name',
        'nickname',
        'host',
        'avatar_url',
        'is_admin',
        'is_moderator',
        'is_bot',
        'is_cat',
        'is_lady',
        'emojis',
        'online_status',
        'url',
        'uri',
        'created_at',
        'updated_at',
        'is_locked',
        'is_silenced',
        'is_suspended',
        'description',
        'location',
        'birthday',
        'fields',
        'followers_count',
        'following_count',
        'notes_count',
        'pinned_note_ids',
        'pinned_notes',
        'pinned_page_id',
        'pinned_page',
        'ff_visibility',
        'is_following',
        'is_follow',
        'is_blocking',
        'is_blocked',
        'is_muted',
        'details',
        'instance',
    )

    def __init__(self, data: UserPayload):
        self.id: str = data['user_id'] if data.get('user_id') else data['id']
        self.name: str = data['username']
        self.nickname: Optional[str] = data.get('name')
        self.host: Optional[str] = data.get('host')
        self.avatar_url: Optional[str] = data.get('avatar_url')
        self.is_admin: bool = bool(data.get('is_admin'))
        self.is_moderator: bool = bool(data.get('is_moderator'))
        self.is_bot: bool = bool(data.get('is_bot'))
        self.is_cat: bool = bool(data.get('is_cat', False))
        self.is_lady: bool = bool(data.get('is_lady', False))
        self.emojis: Optional[List[str]] = data.get('emojis')
        self.online_status = data.get('online_status', None)
        self.url: Optional[str] = data.get('url')
        self.uri: Optional[str] = data.get('uri')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.updated_at = data.get('updated_at')
        self.is_locked = data.get('is_locked', False)
        self.is_silenced = data.get('is_silenced', False)
        self.is_suspended = data.get('is_suspended', False)
        self.description = data.get('description')
        self.location = data.get('location')
        self.birthday = data.get('birthday')
        self.fields = data.get('fields', [])
        self.followers_count = data.get('followers_count', 0)
        self.following_count = data.get('following_count', 0)
        self.notes_count = data.get('notes_count', 0)
        self.pinned_note_ids = data.get('pinned_note_ids', [])
        self.pinned_notes = data.get('pinned_notes', [])
        self.pinned_page_id = data.get('pinned_page_id')
        self.pinned_page = data.get('pinned_page')
        self.ff_visibility: str = data.get('ff_visibility', 'public')
        self.is_following: bool = bool(data.get('is_following', False))
        self.is_follow: bool = bool(data.get('is_follow', False))
        self.is_blocking: bool = bool(data.get('is_blocking', False))
        self.is_blocked: bool = bool(data.get('is_blocked', False))
        self.is_muted: bool = bool(data.get('is_muted', False))
        self.details: RawUserDetails = RawUserDetails(data)
        self.instance: Optional[RawInstance] = RawInstance(
            data['instance']
        ) if data.get('instance') else None
